/*  This file is part of SAIL (https://github.com/smoked-herring/sail)

    Copyright (c) 2020 Dmitry Baryshev

    The MIT License

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.
*/

#include "sail-common.h"
#include "sail.h"
#include "sail-c++.h"

namespace sail
{

class SAIL_HIDDEN write_features::pimpl
{
public:
    pimpl()
        : sail_write_features_c(nullptr)
        , features(0)
        , properties(0)
        , preferred_compression_type(SAIL_COMPRESSION_UNSUPPORTED)
        , compression_min(0)
        , compression_max(0)
        , compression_default(0)
    {}

    const sail_write_features *sail_write_features_c;

    std::map<SailPixelFormat, std::vector<SailPixelFormat>> pixel_formats_mappings;
    int features;
    int properties;
    std::vector<SailCompressionType> compression_types;
    SailCompressionType preferred_compression_type;
    int compression_min;
    int compression_max;
    int compression_default;
};

write_features::write_features(const write_features &wf)
    : write_features()
{
    *this = wf;
}

write_features& write_features::operator=(const write_features &wf)
{
    d->sail_write_features_c = wf.d->sail_write_features_c;

    with_pixel_formats_mappings(wf.pixel_formats_mappings())
        .with_features(wf.features())
        .with_properties(wf.properties())
        .with_compression_types(compression_types())
        .with_preferred_compression_type(wf.preferred_compression_type())
        .with_compression_min(wf.compression_min())
        .with_compression_max(wf.compression_max())
        .with_compression_default(wf.compression_default());

    return *this;
}

write_features::~write_features()
{
    delete d;
}

std::map<SailPixelFormat, std::vector<SailPixelFormat>> write_features::pixel_formats_mappings() const
{
    return d->pixel_formats_mappings;
}

int write_features::features() const
{
    return d->features;
}

int write_features::properties() const
{
    return d->properties;
}

std::vector<SailCompressionType> write_features::compression_types() const
{
    return d->compression_types;
}

SailCompressionType write_features::preferred_compression_type() const
{
    return d->preferred_compression_type;
}

int write_features::compression_min() const
{
    return d->compression_min;
}

int write_features::compression_max() const
{
    return d->compression_max;
}

int write_features::compression_default() const
{
    return d->compression_default;
}

sail_error_t write_features::to_write_options(write_options *swrite_options) const
{
    SAIL_CHECK_WRITE_FEATURES_PTR(d->sail_write_features_c);
    SAIL_CHECK_WRITE_OPTIONS_PTR(swrite_options);

    sail_write_options *sail_write_options;

    SAIL_TRY(sail_alloc_write_options_from_features(d->sail_write_features_c, &sail_write_options));

    *swrite_options = write_options(sail_write_options);

    sail_destroy_write_options(sail_write_options);

    return 0;
}

write_features::write_features()
    : d(new pimpl)
{
}

write_features::write_features(const sail_write_features *wf)
    : write_features()
{
    if (wf == nullptr) {
        SAIL_LOG_ERROR("NULL pointer has been passed to sail::write_features()");
        return;
    }

    d->sail_write_features_c = wf;

    std::map<SailPixelFormat, std::vector<SailPixelFormat>> pixel_formats_mappings;
    sail_pixel_formats_mapping_node *node = wf->pixel_formats_mapping_node;

    while (node != nullptr) {
        std::vector<SailPixelFormat> pixel_formats;
        pixel_formats.reserve(node->output_pixel_formats_length);

        for (int i = 0; i < node->output_pixel_formats_length; i++) {
            pixel_formats.push_back(node->output_pixel_formats[i]);
        }

        pixel_formats_mappings[node->input_pixel_format] = pixel_formats;

        node = node->next;
    }

    std::vector<SailCompressionType> compression_types;

    if (wf->compression_types != nullptr && wf->compression_types_length > 0) {
        compression_types.reserve(wf->compression_types_length);

        for (int i = 0; i < wf->compression_types_length; i++) {
            compression_types.push_back(wf->compression_types[i]);
        }
    }

    with_pixel_formats_mappings(pixel_formats_mappings)
        .with_features(wf->features)
        .with_properties(wf->properties)
        .with_compression_types(compression_types)
        .with_preferred_compression_type(wf->preferred_compression_type)
        .with_compression_min(wf->compression_min)
        .with_compression_max(wf->compression_max)
            .with_compression_default(wf->compression_default);
}

write_features &write_features::with_pixel_formats_mappings(const std::map<SailPixelFormat, std::vector<SailPixelFormat>> &pixel_formats_mappings)
{
    d->pixel_formats_mappings = pixel_formats_mappings;
    return *this;
}

write_features& write_features::with_features(int features)
{
    d->features = features;
    return *this;
}

write_features& write_features::with_properties(int properties)
{
    d->properties = properties;
    return *this;
}

write_features& write_features::with_compression_types(const std::vector<SailCompressionType> &compression_types)
{
    d->compression_types = compression_types;
    return *this;
}

write_features& write_features::with_preferred_compression_type(SailCompressionType preferred_compression_type)
{
    d->preferred_compression_type = preferred_compression_type;
    return *this;
}

write_features& write_features::with_compression_min(int compression_min)
{
    d->compression_min = compression_min;
    return *this;
}

write_features& write_features::with_compression_max(int compression_max)
{
    d->compression_max = compression_max;
    return *this;
}

write_features& write_features::with_compression_default(int compression_default)
{
    d->compression_default = compression_default;
    return *this;
}

const sail_write_features* write_features::sail_write_features_c() const
{
    return d->sail_write_features_c;
}

}
